package proj3;

import java.util.HashMap;
import java.util.Iterator;
import java.util.PriorityQueue;

import proj3.RoomNode.RoomState;

/**
 * As Best First, A*, and Dijkstra's share so much code, it'd be cumbersome to rewrite it.
 * Thus, they all share the base code for utilizing a heap as the open list that keeps the
 * most promising node at the root. The difference comes in the way that the current path 
 * length is updated. The path length is controlled by methods that are left to the
 * individual algorithms.
 */
public abstract class HeapGraphSearch implements GraphSearch {
	private int expanded = 0;
	
	/**
	 * This is what actual goes into the heap, since it implements the Comparable interface.
	 * The ordering is such that, given RoomNodes a & b, a &lt; b iff the estimated path length
	 * by following a is shorter than that of b. I.e., the shortest path is at the minimum of the
	 * heap.
	 */
	protected class RoomInfo implements Comparable<RoomInfo> {
		public int cost;
		public int heuristic;
		public RoomNode room;
		
		public RoomInfo(RoomNode room, int cost, int heuristic) {
			this.room = room;
			this.cost = cost;
			this.heuristic = heuristic;
		}
		
		/**
		 * Returns an integer representing the ordering of two RoomInfos
		 * The RoomInfo with the shortest estimated path length is considered less than the other.
		 * Because of a bug in PriorityQueue, remove() is based on compareTo() as opposed to equals(). [bug #6207984]
		 * As such, when the two paths are the same, 0 is only returned if these represent the same node.
		 * This should not adversely affect the correctness.
		 */
		public int compareTo(RoomInfo n) {
			int f = cost + heuristic;
			int f2 = n.cost + n.heuristic;
			if(f < f2) return -1;
			else if(f == f2) return (this == n) ? 0 : -1;
			else return 1;
		}
	}
	
	public boolean search(RoomNode start, RoomNode end) {
		//The info map allows us to retreive information about the current path to any given node
		HashMap<RoomNode,RoomInfo> info_map = new HashMap<RoomNode,RoomInfo>();
		//The heap is used to store the open list
		PriorityQueue<RoomInfo> heap = new PriorityQueue<RoomInfo>();
		//Initialize the heap by inserting the start node
		RoomInfo info = new RoomInfo(start, 0, heuristic(start,end));
		expanded = 0;
		heap.offer(info);
		info_map.put(start, info);
		
		while(true) { //Keep going until either we find the end or the heap is empty
			if(heap.isEmpty())
				return false; //If the heap is empty here, we can't get to the end
			
			info = heap.poll(); //Grab the most promising node
			expanded++;
			info.room.setState(RoomState.SEEN);
			if(info.room == end)
				break; //We've reached the end
			Iterator<RoomNode> it = info.room.getNeighbors();
			while(it.hasNext()) {
				//More to go, so add all of this node's neighbors to the open list
				RoomNode child = it.next();
				RoomInfo newInfo = new RoomInfo(child,newCost(info),heuristic(child,end));
				if(child.getState() != RoomState.UNSEEN) {
					//Check to see if the new path here is any better
					RoomInfo childInfo = info_map.get(child);
					if(childInfo != null) {
						if(childInfo.compareTo(newInfo) <= 0)
							continue; //Nope, forget this new path
						else
							heap.remove(childInfo); //Yep, we'll add the new path, so get rid of the old
					}
				}
				//Add the node to the open list
				child.setState(RoomState.IN_OPEN_LIST);
				child.setParent(info.room);
				info_map.put(child,newInfo);
				heap.offer(newInfo);
			}
		}
		//Set up the path doubly-linked list, and change the state of those rooms on the path
		RoomNode n = end;
		while(n != start) {
			RoomNode parent = n.getParentPath();
			parent.setNextOnPath(n);
			parent.setState(RoomState.ON_PATH);
			n = parent;
		}
		end.setState(RoomState.ON_PATH);
		
		return true;
	}
	
	public int getLastExpandedNodeCount() { return expanded; }
	
	/**
	 * Returns an estimate of the distance to the goal
	 */
	abstract protected int heuristic(RoomNode start, RoomNode end);
	/**
	 * Returns the cost of the path traveled so far and one additional step.
	 * This abstraction allows us to use the same code for both A* and Best First 
	 */
	abstract protected int newCost(RoomInfo n);
}
