package proj3;

import java.io.PrintStream;
import java.util.Iterator;
import java.util.LinkedList;

import proj3.RoomNode.RoomState;

import base.EasyReader;

/**
 * Maze holds a list of rooms that represent the maze. This also contains a method
 * to parse a maze file and also output a text based version to a file.
 */
public class Maze {
	public static char ROOM_START = '*';
	public static char ROOM_END = 'X';
	public static char ROOM_EMPTY = ' ';
	
	public static char ROOM_CORNER = '+';
	public static char ROOM_HORIZ_WALL = '-';
	public static char ROOM_VERT_WALL = '|';
	public static char ROOM_DOOR = ' ';
	
	protected RoomNode[][] rooms;
	//Room at (x,y) is rooms[x][y]
	protected RoomNode start, end;
	protected int height, width;
	
	public Maze(String filename) {
		parseMaze(filename);
	}
	
	public Maze(RoomNode[][] rooms, RoomNode start, RoomNode end, int width, int height) {
		this.rooms = rooms;
		this.start = start;
		this.end = end;
		this.width = width;
		this.height = height;
	}
	
	public RoomNode[][] getRooms() { return rooms; }
	public RoomNode getStart() { return start; }
	public RoomNode getEnd() { return end; }
	public int getHeight() { return height; }
	public int getWidth() { return width; }
	
	/**
	 * This method should be called before starting any search, as the searching relies
	 * on having all the rooms start in the UNSEEN state.
	 */
	public void reset() {
		for(RoomNode[] row : rooms) {
			for(RoomNode n : row) {
				n.setState(RoomState.UNSEEN);
				n.setNextOnPath(null);
				n.setParent(null);
			}
		}
	}
	
	private void parseMaze(String filename) {
		EasyReader in = new EasyReader(filename);
		//Parse it
		width = in.readInt();
		height = in.readInt();
		in.readLine(); //Toss out newline
		
		rooms = new RoomNode[width][height];
		for(int y = 0; y < height; y++) {
			for(int x = 0; x < width; x++) {
				rooms[x][y] = new RoomNode(x,y);
			}
		}
		
		//Skip top line, since it's always the same
		in.readLine();
		//Process each line
		for(int y = 0; y < height; y++) {
			char c = 0;
			in.readChar(); //Read left wall and toss it out
			for(int x = 0; x < width; x++) {
				c = in.readChar(); //This room
				if(c == ROOM_START) {
					if(start != null)
						throw new BadMazeException();
					start = rooms[x][y];
				} else if(c == ROOM_END) {
					if(end != null)
						throw new BadMazeException();
					end = rooms[x][y];
				} else if(c == '\n') {
					break;
				}
				c = in.readChar(); //Wall to the right of this room
				if(c == ROOM_DOOR && x < width-1) {
					RoomNode.markNeighbors(rooms[x][y],rooms[x+1][y]);
				}
			}
			if(c != '\n')
				in.readLine(); //Read the newline character and toss it
			in.readChar(); //Read left wall (corner, actually) and toss it
			if(y < height - 1) {
				for(int x = 0; x < width; x++) {
					c = in.readChar();
					if(c == ROOM_DOOR) {
						RoomNode.markNeighbors(rooms[x][y],rooms[x][y+1]);
					} else if(c == '\n') {
						break;
					}
					in.readChar(); //Read corner and toss it
				}
			}
			if(c != '\n')
				in.readLine(); //Toss the newline character
		}
		if(start == null || end == null)
			throw new BadMazeException();
		removeEmptyRooms();
	}
	
	private void removeEmptyRooms() {
		//This is an adaptation of generic graph search that marks all the nodes that can be reached.
		LinkedList<RoomNode> list = new LinkedList<RoomNode>();
		reset();
		start.setState(RoomState.SEEN);
		list.add(start);
		int expanded = 0;
		while(!list.isEmpty()) {
			Iterator<RoomNode> it = list.poll().getNeighbors();
			expanded++;
			while(it.hasNext()) {
				RoomNode n = it.next();
				if(n.getState() == RoomState.UNSEEN) {
					n.setState(RoomState.SEEN);
					list.offer(n);
				}
			}
		}
		for(int y = 0; y < height; y++) {
			for(int x = 0; x < width; x++) {
				if(rooms[x][y].getState() == RoomState.UNSEEN) {
					if(rooms[x][y] == end)
						throw new BadMazeException();
					rooms[x][y] = null;
				}
				else
					rooms[x][y].setState(RoomState.UNSEEN); //Need to reset the state
			}
		}
	}
	
	public void printMaze(PrintStream out) {
		//Print top wall
		for(int x = 0; x < width; x++) {
			if(rooms[x][0] != null || (x > 0 && rooms[x-1][0] != null))
				out.print(ROOM_CORNER);
			else
				out.print(" ");
			if(rooms[x][0] != null)
				out.print(ROOM_HORIZ_WALL);
			else
				out.print(" ");
		}
		if(rooms[width-1][0] != null)
			out.println(ROOM_CORNER);
		else
			out.println();
		
		for(int y = 0; y < height; y++) {
			//Print out the row with the rooms in it
			if(rooms[0][y] != null)
				out.print(ROOM_VERT_WALL);
			else
				out.print(" ");
			for(int x = 0; x < width; x++) {
				RoomNode n = rooms[x][y];
				if(n == null) {
					if(x < width-1 && rooms[x+1][y] != null)
						out.print(" |");
					else
						out.print("  ");
					continue;
				}
				//Print the room itself
				if(n == start) {
					out.print(ROOM_START);
				} else if(n == end) {
					out.print(ROOM_END);
				} else {
					out.print(n.getState().getMapChar());
				}
				//Print the door
				if(x < width-1 && n.isAdjacentTo(rooms[x+1][y])) {
					out.print(ROOM_DOOR);
				} else {
					out.print(ROOM_VERT_WALL);
				}
			}
			out.println();
			
			//Print out the wall in-between the row of rooms
			if(rooms[0][y] != null || (y < height-1 && rooms[0][y+1] != null))
				out.print(ROOM_CORNER);
			else
				out.print(" ");
			for(int x = 0; x < width; x++) {
				//Print the door
				if(rooms[x][y] == null) {
					if(y < height-1 && rooms[x][y+1] != null)
						out.print("-+");
					else if(x < width-1 && (rooms[x+1][y] != null ||
							(y < height-1 && rooms[x+1][y+1] != null)))
						out.print(" +");
					else
						out.print("  ");
					continue;
				}
				if(y < height-1 && rooms[x][y].isAdjacentTo(rooms[x][y+1])) {
					out.print(ROOM_DOOR);
				} else {
					out.print(ROOM_HORIZ_WALL);
				}
				//Print the corner
				if(x < width-1 && y < height-1 &&
						rooms[x][y].isAdjacentTo(rooms[x+1][y]) &&
						(rooms[x][y+1] != null && rooms[x][y+1].isAdjacentTo(rooms[x+1][y+1])) &&
						rooms[x][y].isAdjacentTo(rooms[x][y+1]) &&
						(rooms[x+1][y] != null && rooms[x+1][y].isAdjacentTo(rooms[x+1][y+1]))) {
					//This checks to see if there's a wide-open empty room spanning multiple nodes
					out.print(ROOM_EMPTY);
				} else {
					out.print(ROOM_CORNER);
				}
			}
			out.println();
		}
	}
	
	public class BadMazeException extends RuntimeException {
		
	}
}
