package proj3;

import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.List;

import javax.swing.UIManager;

import proj3.extras.StepAStarSearch;
import proj3.extras.StepBestFirstSearch;
import proj3.extras.StepBreadthFirstSearch;
import proj3.extras.StepDepthFirstSearch;
import proj3.mazegen.RandomMaze;

import view.MazeRunnerWindow;

/**
 * MazeRunner is the main application.
 */
public class MazeRunner {
	static final String ALG_DEPTH_FIRST = "dfs";
	static final String ALG_BREADTH_FIRST = "bfs";
	static final String ALG_BEST_FIRST = "best";
	static final String ALG_A_STAR = "astar";
	static final String ALG_STEP_DEPTH_FIRST = "sdfs";
	static final String ALG_STEP_BREADTH_FIRST = "sbfs";
	static final String ALG_STEP_BEST_FIRST = "sbest";
	static final String ALG_STEP_A_STAR = "sastar";
	
	static final String ARG_VISUAL = "-vis";
	static final String ARG_INPUT = "-input";
	static final String ARG_OUTPUT = "-final";
	static final String ARG_WIDTH = "-w";
	static final String ARG_HEIGHT = "-h";
	static final String ARG_CYCLES = "-c";
	
	public static void main(String[] args) {
		if(args.length < 1) {
			System.err.println("Command line arguments not recognized.");
			printUsage();
			return;
		}
		
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch(Exception e) {}
		System.setProperty("swing.aatext", "true"); 
		
		GraphSearch gs = null;
		String file_in = null;
		String file_out = null;
		boolean visual = false;
		int width = 15;
		int height = 15;
		int cycles = 15;
		
		//Process args
		String alg = args[0];
		if(alg.equals(ALG_DEPTH_FIRST)) {
			gs = new DepthFirstSearch();
		} else if(alg.equals(ALG_BREADTH_FIRST)) {
			gs = new BreadthFirstSearch();
		} else if(alg.equals(ALG_BEST_FIRST)) {
			gs = new BestFirstSearch();
		} else if(alg.equals(ALG_A_STAR)) {
			gs = new AStarSearch();
		} else if(alg.equals(ALG_STEP_DEPTH_FIRST)) {
			gs = new StepDepthFirstSearch();
		} else if(alg.equals(ALG_STEP_BREADTH_FIRST)) {
			gs = new StepBreadthFirstSearch();
		} else if(alg.equals(ALG_STEP_BEST_FIRST)) {
			gs = new StepBestFirstSearch();
		} else if(alg.equals(ALG_STEP_A_STAR)) {
			gs = new StepAStarSearch();
		} else {
			System.err.println("Command line arguments not recognized.");
			printUsage();
			return;
		}
		
		for(int i = 1; i < args.length;) {
			if(args[i].equals(ARG_INPUT)) {
				file_in = args[i+1];
				i += 2;
			} else if(args[i].equals(ARG_OUTPUT)) {
				file_out = args[i+1];
				i += 2;
			} else if(args[i].equals(ARG_VISUAL)) {
				visual = true;
				i++;
			} else if(args[i].startsWith(ARG_WIDTH)) {
				width = Integer.parseInt(args[i].substring(2));
				i++;
			} else if(args[i].startsWith(ARG_HEIGHT)) {
				height = Integer.parseInt(args[i].substring(2));
				i++;
			} else if(args[i].startsWith(ARG_CYCLES)) {
				cycles = Integer.parseInt(args[i].substring(2));
				i++;
			} else {
				i++;
			}
		}
		Maze m;
		if(file_in == null) {
			/*System.err.println("Command line arguments not recognized.");
			printUsage();
			return;*/
			m = RandomMaze.generateRandomMaze(width,height,cycles);
		} else {
			//Construct the maze
			m = new Maze(file_in);
		}
		//m.printMaze(System.out);
		
		//Solve it
		gs.search(m.getStart(), m.getEnd());
		List<RoomNode> path = m.getStart().getPath();
		
		//Print output
		System.out.println(file_in);
		System.out.println(gs);
		printPath(path);
		System.out.println(path.size());
		System.out.println(gs.getLastExpandedNodeCount());
		System.out.println();
		
		//Print to file
		if(file_out != null) {
			try {
				m.printMaze(new PrintStream(file_out));
			} catch(FileNotFoundException e) {
				System.err.println("Couldn't create output file: " + file_out);
			}
		}
		
		if(visual) {
			new MazeRunnerWindow(m, gs); //Looks ugly, but it pops up the window, and then we're done here
		}
	}
	
	public static void printPath(List<RoomNode> path) {
		for(RoomNode n : path) {
			System.out.print("(" + n.getX() + "," + n.getY() + ") ");
		}
		System.out.println();
	}
	
	public static void printUsage() {
		System.out.println("Usage: java MazeRunner alg [-input infile] [-final picfile] [-wX -hX -cX] [-vis]");
		System.out.println();
		System.out.println("alg must be supplied, and describes the algorithm to use. It is one of the following:");
		System.out.println("\t" + ALG_DEPTH_FIRST + "\tDepth-First Search");
		System.out.println("\t" + ALG_BREADTH_FIRST + "\tBreadth-First Search");
		System.out.println("\t" + ALG_BEST_FIRST + "\tBest-First Search");
		System.out.println("\t" + ALG_A_STAR + "\tA* Search");
		System.out.println();
		System.out.println("infile is a file that specifies the maze.");
		System.out.println("picfile is an optional argument that specifies a filename to print a picture of the completed maze to.");
		System.out.println("If infile is not specified, a random maze will be generated.");
		System.out.println("-w -h -c specify the width, height, and cycle count, respectively, of the random maze.");
		System.out.println("When -vis is given, the program will run in a GUI mode.");
	}
}
