package proj3.extras;

import java.util.HashMap;
import java.util.Iterator;
import java.util.PriorityQueue;

import proj3.HeapGraphSearch;
import proj3.RoomNode;
import proj3.RoomNode.RoomState;

/**
 * As Best First, A*, and Dijkstra's share so much code, it'd be cumbersome to rewrite it.
 * Thus, they all share the base code for utilizing a heap as the open list that keeps the
 * most promising node at the root. The difference comes in the way that the current path 
 * length is updated. The path length is controlled by methods that are left to the
 * individual algorithms.
 */
public abstract class StepHeapGraphSearch implements StepGraphSearch {
	HashMap<RoomNode,RoomInfo> info_map;
	PriorityQueue<RoomInfo> heap;
	private RoomNode start, end;
	private int expanded = 0;
	
	public StepHeapGraphSearch() {
		info_map = new HashMap<RoomNode,RoomInfo>();
		heap = new PriorityQueue<RoomInfo>();
	}
	
	/**
	 * @see HeapGraphSearch.RoomInfo
	 */
	protected class RoomInfo implements Comparable<RoomInfo> {
		public int cost;
		public int heuristic;
		public RoomNode room;
		
		public RoomInfo(RoomNode room, int cost, int heuristic) {
			this.room = room;
			this.cost = cost;
			this.heuristic = heuristic;
		}
		
		/**
		 * Returns an integer representing the ordering of two RoomInfos
		 * The RoomInfo with the shortest estimated path length is considered less than the other.
		 * Because of a bug in PriorityQueue, remove() is based on compareTo() as opposed to equals(). [bug #6207984]
		 * As such, when the two paths are the same, 0 is only returned if these represent the same node.
		 * This should not adversely affect the correctness.
		 */
		public int compareTo(RoomInfo n) {
			int f = cost + heuristic;
			int f2 = n.cost + n.heuristic;
			if(f < f2) return -1;
			else if(f == f2) return (this == n) ? 0 : -1;
			else return 1;
		}
	}
	
	public void setStartAndEnd(RoomNode start, RoomNode end) {
		this.start = start;
		this.end = end;
		info_map.clear();
		heap.clear();
		RoomInfo info = new RoomInfo(start, 0, heuristic(start,end));
		expanded = 0;
		heap.offer(info);
		info_map.put(start, info);
	}
	
	/**
	 * @see HeapGraphSearch#search(RoomNode, RoomNode)
	 */
	public boolean advance() {
		if(heap.isEmpty())
			return true;
		
		RoomInfo info = heap.poll();
		expanded++;
		if(info.room == end) {
			RoomNode n = end;
			while(n != start) {
				RoomNode parent = n.getParentPath();
				parent.setNextOnPath(n);
				parent.setState(RoomState.ON_PATH);
				n = parent;
			}
			end.setState(RoomState.ON_PATH);
			info_map.clear();
			heap.clear();
			return true;
		}
		
		info.room.setState(RoomState.SEEN);
		Iterator<RoomNode> it = info.room.getNeighbors();
		while(it.hasNext()) {
			RoomNode child = it.next();
			RoomInfo newInfo = new RoomInfo(child,newCost(info),heuristic(child,end));
			if(child.getState() != RoomState.UNSEEN) {
				RoomInfo childInfo = info_map.get(child);
				if(childInfo != null) {
					if(childInfo.compareTo(newInfo) <= 0)
						continue;
					else
						heap.remove(childInfo);
				}
			}
			child.setState(RoomState.IN_OPEN_LIST);
			child.setParent(info.room);
			info_map.put(child,newInfo);
			heap.offer(newInfo);
		}
		
		return false;
	}
	
	public boolean search(RoomNode start, RoomNode end) {
		setStartAndEnd(start, end);
		while(!advance());
		return true;
	}
	
	public int getLastExpandedNodeCount() { return expanded; }	
	
	abstract protected int heuristic(RoomNode start, RoomNode end);
	abstract protected int newCost(RoomInfo n);
}
