package proj3.mazegen;

import java.util.Random;
import proj3.RoomNode;
import proj3.Maze;

import static proj3.mazegen.Edge.*;

/**
 * This class contains static methods for generating a random maze.
 *
 */

public class RandomMaze {
	static Vertex[][] vertices;
	static Edge[] edges;
	

	/* Random number generator */
	static final Random rng = new Random();
	
	/** 
	 * This method performs disjoin union on two vertices.
	 * @param a - the first vertex.
	 * @param b - the second vertex.
	 * 
	 */
	public static void link(Vertex a, Vertex b) {
		if(a.getRank() == b.getRank()) {
			a.setParent(b);
			b.setRank(b.getRank() + 1);
		} else if(a.getRank() > b.getRank()) {
			b.setParent(a);
		} else {
			a.setParent(b);
		}
	}
	
	
	/**
	 * 
	 * @param a - a vertex.
	 * @return - the vertex representing the set that a is in.
	 */
	public static Vertex find(Vertex a) {
		if(a != a.getParent()) {
			a.setParent(find(a.getParent())); //Path compression
		}
		return a.getParent();
	}
	
	
	/**
	 * This method generates a random maze based on the given parameters.
	 * @param width - the width of the maze to be generated.
	 * @param height - the height of the maze to be generated.
	 * @param cycles - the # of cycles to create after the complete maze is generated
	 * @return - a new maze.
	 */
	public static Maze generateRandomMaze(int width, int height, int cycles) {
		vertices = new Vertex[width][height];
		edges = new Edge[width * (height - 1) + height * (width - 1)];
		
		// Initialize vertices
		for(int y = 0; y < height; y++) {
			for(int x = 0; x < width; x++) {
				vertices[x][y] = new Vertex(0,x,y);
			}
		}
		
		// Initialize edges
		
		// Initialize the vertical edges.
		for(int y = 0; y < (height - 1); y++) {
			for(int x = 0; x < width; x++) {
				edges[width * y + x] = new Edge(x,y,EDGE_VERTICAL,true);
			}
		}
		
		// Initialize the horizontal edges
		int offset = width * (height - 1);
		for(int y = 0; y < height; y++) {
			for(int x = 0; x < (width - 1); x++) {
				edges[offset + (width - 1) * y + x] = new Edge(x,y,EDGE_HORIZONTAL,true);
			}
		}

		// Randomize edges
		for(int i = edges.length-1; i > 0; i--) {
			Edge tmp = edges[i];
			int j = (int) (rng.nextInt(i));
			edges[i] = edges[j];
			edges[j] = tmp;
		}
		
		// Create maze
		for(int i = 0; i < edges.length; i++) {
			int x = edges[i].getX();
			int y = edges[i].getY();
			Vertex a = find(vertices[x][y]);
			Vertex b;
			if(edges[i].getPos() == EDGE_VERTICAL) {
				b = find(vertices[x][y + 1]);
			} else {
				b = find(vertices[x + 1][y]);
			}

			if(a != b) {
				link(a, b);
				edges[i].setWall(false);
			}
		}
		
		// Randomly remove additional walls
		for(int i = 0; (i < edges.length) && cycles > 0; i++) {
			if(edges[i].isWall()) {
				edges[i].setWall(false);
				cycles--;
			}
		}
		
		// Initialize an array of RoomNodes that will be passed to the Maze constructor
		RoomNode[][] nodes = new RoomNode[width][height];
		for(int y = 0; y < height; y++) {
			for(int x = 0; x < width; x++) {
				nodes[x][y] = new RoomNode(x, y);
			}
		}
		
		// Loop through the edges to mark adjacent vertices as neighbors
		for(int i = 0; i < edges.length; i++) {
			Edge e = edges[i];
			if(!e.isWall()) {
				int x = e.getX();
				int y = e.getY();
				if(e.getPos() == EDGE_VERTICAL) {
					RoomNode.markNeighbors(nodes[x][y],nodes[x][y+1]);
				} else { // It is a horizontal edge
					RoomNode.markNeighbors(nodes[x][y],nodes[x+1][y]);
				}
			}
		}
		
		Maze m = new Maze(nodes, nodes[0][0], nodes[width-1][height-1], width, height);
		return m;
	}
}
