package view;

import java.awt.BorderLayout;
import java.io.File;
import java.io.PrintStream;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.WindowConstants;
import javax.swing.filechooser.FileFilter;

import proj3.GraphSearch;
import proj3.Maze;
import proj3.RoomNode.RoomState;
import proj3.extras.StepGraphSearch;
import proj3.mazegen.RandomMaze;

/**
 * This class is the main window for the GUI. It automatically shows itself when the constructor is called.
 */
public class MazeRunnerWindow extends JFrame {
	private JScrollPane pane;
	private MazeView m;
	private GraphSearch gs;
	private StatusBar statusBar;
	
	public JFileChooser fileChooser;
	
	public MazeRunnerWindow(Maze maze, GraphSearch gs) {
		super("MazeRunner");
		
		pane = new JScrollPane();
		m = new MazeView(maze);
		statusBar = new StatusBar();
		
		MenuHandler mh = new MenuHandler(this);
		JMenuBar menu = mh.getMenu(gs);
		setJMenuBar(menu);
		setAlgorithm(gs);
		
		pane.setViewportView(m);
		pane.setMaximumSize(m.getSize());
		
		getContentPane().add(pane, BorderLayout.CENTER);
		getContentPane().add(statusBar, BorderLayout.SOUTH);
		pack();
		setVisible(true);
		repaint();
		
		setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
		fileChooser = new JFileChooser();
		fileChooser.setFileFilter(new FileFilter() {
			public boolean accept(File f) {
				return f.isDirectory() || (f.getName().toLowerCase().endsWith(".txt") ? true : false);
			}
			
			public String getDescription() {
		        return "Text Documents (*.txt)";
		    }
		});
	}
	
	public Maze getMaze() { return m.getMaze(); }
	
	// These methods allow the MenuHandler to control the Maze
	
	public void quit() {
		dispose();
		System.exit(0);
	}
	public void advanceSearch() {
		if(gs instanceof StepGraphSearch) {
			((StepGraphSearch)gs).advance();
		} else {
			//Shouldn't even be able to get here, since the menu item is grayed out
			JOptionPane.showMessageDialog(null, gs + " doesn't support Next.","Uh-oh",JOptionPane.ERROR_MESSAGE);
			System.err.println("advance() called on non Step-by-step algorithm.");
		}
		repaint();
	}
	public void finishSearch() {
		if(gs instanceof StepGraphSearch) {
			StepGraphSearch sgs = (StepGraphSearch)gs;
			while(!sgs.advance());
		} else {
			Maze maze = m.getMaze();
			gs.search(maze.getStart(), maze.getEnd());
		}
		repaint();
	}
	public void setAlgorithm(GraphSearch gs) {
		Maze maze = m.getMaze();
		if(this.gs != null && this.gs.getClass() != gs.getClass()) {
			resetMaze();
		}
		this.gs = gs;
		repaint(); //Update status bar
		if(gs instanceof StepGraphSearch) {
			StepGraphSearch sgs = (StepGraphSearch)gs;
			sgs.setStartAndEnd(maze.getStart(), maze.getEnd());
		}
	}
	public void resetMaze() {
		m.getMaze().reset();
		if(gs instanceof StepGraphSearch) {
			Maze maze = m.getMaze();
			((StepGraphSearch)gs).setStartAndEnd(maze.getStart(),maze.getEnd());
		}
		repaint();
	}
	public void openMaze() {
		if(fileChooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
			Maze maze = new Maze(fileChooser.getSelectedFile().getPath());
			m.setMaze(maze);
			pack();
			repaint();
		}
	}
	public void printMaze() {
		if(fileChooser.showSaveDialog(null) == JFileChooser.APPROVE_OPTION) {
			try {
				m.getMaze().printMaze(new PrintStream(fileChooser.getSelectedFile().getPath()));
			} catch(Exception e) {
				JOptionPane.showMessageDialog(null,
						"Error writing to file",
						"File I/O Error",
						JOptionPane.ERROR_MESSAGE);
			}
		}
	}
	public void repaint() {
		super.repaint();
		JMenu menu = getJMenuBar().getMenu(MenuHandler.MENUI_MAZE);
		//Enable/disable the algorithm control menu options based on whether the search is complete already
		if(m.getMaze().getStart().getState() == RoomState.ON_PATH) {
			menu.getMenuComponent(0).setEnabled(false);
			menu.getMenuComponent(1).setEnabled(false);
		} else {
			menu.getMenuComponent(0).setEnabled(gs instanceof StepGraphSearch);
			menu.getMenuComponent(1).setEnabled(true);
		}
		statusBar.setNodeCount(gs.getLastExpandedNodeCount());
		statusBar.setAlgorithm(gs);
	}

	public void setShowRoomColors(boolean b) {
		RoomView.setShowRoomColors(b);
		repaint();
	}

	public void setShowCorners(boolean b) {
		RoomView.setShowCorners(b);
		repaint();
	}

	public void randomMaze(int width, int height, int cycles) {
		Maze maze = RandomMaze.generateRandomMaze(width, height, cycles);
		m.setMaze(maze);
		pack();
		repaint();
	}
}
