package view;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.SystemColor;
import java.util.Iterator;

import proj3.RoomNode;
import proj3.RoomNode.RoomState;

/**
 * This class contains static methods for drawing RoomNodes to a MazeView
 */
public class RoomView {
	static final public int HEIGHT = 20;
	static final public int WIDTH = 20;
	
	//Flag-based representation of all combinations of walls possible 
	static final public int ROOMVIEW_WESTWALL = 0x08;
	static final public int ROOMVIEW_NORTHWALL = 0x04;
	static final public int ROOMVIEW_SOUTHWALL = 0x02;
	static final public int ROOMVIEW_EASTWALL = 0x01;
	
	static final public Color WALL_COLOR = Color.BLACK;
	static final public Color PATH_COLOR = Color.BLACK;
	static final public Color ROOM_COLOR = SystemColor.window;
	
	static private boolean showRoomColors = true;
	static private boolean showCorners = true;
	
	/**
	 * 
	 * @param g The Graphics object to draw onto
	 * @param room The Room to draw
	 * @param dbl Flags for which walls to draw double-thick
	 */
	public static void paintRoom(Graphics g, RoomNode room, int dbl) {
		if(room == null)
			return;
		if(showRoomColors)
			g.setColor(room.getState().getColor());
		else
			g.setColor(ROOM_COLOR);
		
		int left = room.getX()*WIDTH;
		int top = room.getY()*HEIGHT;
		
		//Draw the background
		g.fillRect(left,top,WIDTH,HEIGHT);
		if(room.getState() == RoomState.ON_PATH) {
			//Show the path
			RoomNode next = room.getNextOnPath();
			if(next != null) {
				int left2 = next.getX()*WIDTH;
				int top2 = next.getY()*HEIGHT;
				g.setColor(PATH_COLOR);
				g.drawLine(left + WIDTH/2, top + WIDTH/2, left2 + WIDTH/2, top2 + WIDTH/2);
			}
			next = room.getParentPath();
			if(next != null) {
				int left2 = next.getX()*WIDTH;
				int top2 = next.getY()*HEIGHT;
				g.setColor(PATH_COLOR);
				g.drawLine(left + WIDTH/2, top + WIDTH/2, left2 + WIDTH/2, top2 + WIDTH/2);
			}
		}
		
		//Draw the walls
		g.setColor(WALL_COLOR);
		int walls = getWalls(room);
		if((walls & ROOMVIEW_EASTWALL) > 0) {
			g.drawLine(left+WIDTH-1, top, left+WIDTH-1, top+HEIGHT-1);
		} else if(showCorners) {
			g.drawLine(left+WIDTH-1, top, left+WIDTH-1, top+1);
			g.drawLine(left+WIDTH-1, top+HEIGHT-2, left+WIDTH-1, top+HEIGHT-1);
		}
		if((walls & ROOMVIEW_NORTHWALL) > 0) {
			g.drawLine(left, top, left+WIDTH-1, top);
		} else if(showCorners) {
			g.drawLine(left, top, left+1, top);
			g.drawLine(left+WIDTH-2, top, left+WIDTH-1, top);
		}
		if((walls & ROOMVIEW_SOUTHWALL) > 0) {
			g.drawLine(left, top+HEIGHT-1, left+WIDTH-1, top+HEIGHT-1);
		} else if(showCorners) {
			g.drawLine(left, top+HEIGHT-1, left+1, top+HEIGHT-1);
			g.drawLine(left+WIDTH-2, top+HEIGHT-1, left+WIDTH-1, top+HEIGHT-1);
		}
		if((walls & ROOMVIEW_WESTWALL) > 0) {
			g.drawLine(left, top, left, top+HEIGHT-1);
		} else if(showCorners) {
			g.drawLine(left, top, left, top+1);
			g.drawLine(left, top+HEIGHT-2, left, top+HEIGHT-1);
		}
		if((dbl & ROOMVIEW_EASTWALL) > 0)
			g.drawLine(left+WIDTH, top, left+WIDTH, top+HEIGHT-1);
		if((dbl & ROOMVIEW_NORTHWALL) > 0)
			g.drawLine(left, top-1, left+WIDTH-1, top-1);
		if((dbl & ROOMVIEW_SOUTHWALL) > 0)
			g.drawLine(left, top+HEIGHT, left+WIDTH-1, top+HEIGHT);
		if((dbl & ROOMVIEW_WESTWALL) > 0)
			g.drawLine(left-1, top, left-1, top+HEIGHT-1);
	}
	
	/**
	 * Finds what walls the room has
	 */
	private static int getWalls(RoomNode room) {
		int walls = 0xFF;
		Iterator<RoomNode> it = room.getNeighbors();
		while(it.hasNext()) {
			RoomNode n = it.next();
			int deltaX = n.getX() - room.getX();
			int deltaY = n.getY() - room.getY();
			if(deltaX == 1) {
				walls &= ~ROOMVIEW_EASTWALL;
			} else if(deltaX == -1) {
				walls &= ~ROOMVIEW_WESTWALL;
			} else if(deltaY == 1) {
				walls &= ~ROOMVIEW_SOUTHWALL;
			} else if(deltaY == -1) {
				walls &= ~ROOMVIEW_NORTHWALL;
			}
		}
		return walls;
	}
	
	public static void setShowRoomColors(boolean b) { showRoomColors = b; }
	public static void setShowCorners(boolean b) { showCorners = b; }
}
